// <copyright file="IIterativeSolverSetup.cs" company="Math.NET">
// Math.NET Numerics, part of the Math.NET Project
// http://numerics.mathdotnet.com
// http://github.com/mathnet/mathnet-numerics
// http://mathnetnumerics.codeplex.com
//
// Copyright (c) 2009-2014 Math.NET
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following
// conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
// </copyright>

using System;

namespace MathNet.Numerics.LinearAlgebra.Solvers
{
    /// <summary>
    /// Defines the interface for objects that can create an iterative solver with
    /// specific settings. This interface is used to pass iterative solver creation 
    /// setup information around.
    /// </summary>
    public interface IIterativeSolverSetup<T> where T : struct, IEquatable<T>, IFormattable
    {
        /// <summary>
        /// Gets the type of the solver that will be created by this setup object.
        /// </summary>
        Type SolverType { get; }

        /// <summary>
        /// Gets type of preconditioner, if any, that will be created by this setup object.
        /// </summary>
        Type PreconditionerType { get; }

        /// <summary>
        /// Creates the iterative solver to be used.
        /// </summary>
        IIterativeSolver<T> CreateSolver();

        /// <summary>
        /// Creates the preconditioner to be used by default (can be overwritten).
        /// </summary>
        IPreconditioner<T> CreatePreconditioner();

        /// <summary>
        /// Gets the relative speed of the solver. 
        /// </summary>
        /// <value>Returns a value between 0 and 1, inclusive.</value>
        double SolutionSpeed { get; }

        /// <summary>
        /// Gets the relative reliability of the solver.
        /// </summary>
        /// <value>Returns a value between 0 and 1 inclusive.</value>
        double Reliability { get; }
    }
}
